/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.processor.processor;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Stream;

import jakarta.validation.ConstraintViolationException;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.context.ApplicationEventPublisher;

import de.ozgcloud.apilib.vorgang.OzgCloudVorgangTestFactory;
import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandCreatedEvent;
import de.ozgcloud.command.CommandFailedEvent;
import de.ozgcloud.command.CommandTestFactory;
import de.ozgcloud.command.VorgangCreatedEvent;
import de.ozgcloud.processor.result.ProcessorTechnicalException;
import de.ozgcloud.processor.result.ResultService;
import de.ozgcloud.processor.vorgang.Vorgang;
import de.ozgcloud.processor.vorgang.VorgangId;
import de.ozgcloud.processor.vorgang.VorgangService;
import de.ozgcloud.processor.vorgang.VorgangTestFactory;
import reactor.core.publisher.Mono;

class ProcessorEventListenerTest {

	@Spy
	@InjectMocks
	private ProcessorEventListener vorgangEventListener;

	@Mock
	private VorgangService vorgangService;
	@Mock
	private ProcessorService processorService;
	@Mock
	private ResultService resultService;
	@Mock
	private ApplicationEventPublisher publisher;

	@DisplayName("Trigger new vorgang processor")
	@Nested
	class TestTriggerNewVorgangProcessor {

		@Mock
		private ProcessorTechnicalException exception;

		@Captor
		private ArgumentCaptor<ProcessorResult> resultCaptor;

		private static final VorgangCreatedEvent EVENT = new VorgangCreatedEvent(OzgCloudVorgangTestFactory.ID.toString());

		private Vorgang vorgang = VorgangTestFactory.create();

		@BeforeEach
		void init() {
			when(vorgangService.getVorgang(any())).thenReturn(vorgang);

			var result = ProcessorResultTestFactory.createBuilder().vorgangId(null).build();
			when(processorService.processVorgang(any())).thenReturn(Stream.of(Mono.just(result)));
		}

		@Test
		void shouldGetVorgang() {
			var expectedVorgangId = VorgangId.from(OzgCloudVorgangTestFactory.ID.toString());

			vorgangEventListener.triggerNewVorgangProcessors(EVENT);

			verify(vorgangService).getVorgang(expectedVorgangId);
		}

		@Test
		void shouldCallProcessVorgang() {
			vorgangEventListener.triggerNewVorgangProcessors(EVENT);

			verify(processorService).processVorgang(vorgang);
		}

		@Test
		void shouldSetVorgangId() {
			vorgangEventListener.triggerNewVorgangProcessors(EVENT);

			verify(resultService).processResult(resultCaptor.capture());
			assertThat(resultCaptor.getValue().getVorgangId()).isEqualTo(VorgangTestFactory.ID);
		}

		@Test
		void shouldCallConsumeException() {
			when(vorgangService.getVorgang(any())).thenReturn(VorgangTestFactory.create());
			when(processorService.processVorgang(any())).thenReturn(Stream.of(Mono.error(exception)));

			vorgangEventListener.triggerNewVorgangProcessors(EVENT);

			verify(resultService).processError(exception, VorgangTestFactory.ID.toString());
		}

		@Test
		void shouldCallConsumeUnexpectedExceptionType() {
			when(vorgangService.getVorgang(any())).thenReturn(VorgangTestFactory.create());
			when(processorService.processVorgang(any())).thenReturn(Stream.of(Mono.error(new Exception())));

			vorgangEventListener.triggerNewVorgangProcessors(EVENT);

			verify(resultService).processError(any(Exception.class), eq(VorgangTestFactory.ID.toString()));
		}
	}

	@DisplayName("On command processVorgang")
	@Nested
	class TestOnCommandProcessVorgang {

		@Mock
		private Throwable exception;
		@Captor
		private ArgumentCaptor<CommandFailedEvent> commandFailedEventCaptor;
		@Captor
		private ArgumentCaptor<VorgangProcessedEvent> commandSuccessEventCaptor;

		private final Vorgang vorgang = VorgangTestFactory.create();
		private final Collection<String> processorNames = List.of("dummy1", "dummy2");
		private final Command command = CommandTestFactory.createBuilder()
				.bodyObject(Map.of(ProcessorEventListener.COMMAND_PROCESSOR_NAMES_KEY, processorNames)).build();
		private final CommandCreatedEvent event = new CommandCreatedEvent(command);

		@DisplayName("process flow")
		@Nested
		class TestProcessFlow {

			private final Mono<ProcessorResult> processorResults = Mono.just(ProcessorResultTestFactory.create());
			private final Stream<Mono<ProcessorResult>> monoStream = Stream.of(processorResults);

			@BeforeEach
			void init() {
				when(vorgangService.getVorgang(any())).thenReturn(vorgang);
				when(processorService.processVorgang(any(), any())).thenReturn(monoStream);
			}

			@Test
			void shouldGetVorgang() {
				vorgangEventListener.onCommandProcessVorgang(event);

				verify(vorgangService).getVorgang(VorgangId.from(CommandTestFactory.VORGANG_ID));
			}

			@Test
			void shouldCallService() {
				vorgangEventListener.onCommandProcessVorgang(event);

				verify(processorService).processVorgang(vorgang, processorNames);
			}

			@Test
			void shouldProcessResult() {
				vorgangEventListener.onCommandProcessVorgang(event);

				verify(vorgangEventListener).processResult(any(), eq(CommandTestFactory.VORGANG_ID));
			}

			@Test
			void shouldPublishCommandProcessedEvent() {
				vorgangEventListener.onCommandProcessVorgang(event);

				verify(vorgangEventListener).publishCommandProcessedEvent(event.getSource());
			}
		}

		@DisplayName("get processorNames")
		@Nested
		class TestGetProcessorNames {

			@Test
			void shouldReturnSingletonList() {
				var processorNames = vorgangEventListener.getProcessorNames(createCommand("singleName"));

				assertThat(processorNames).isInstanceOf(Collection.class).containsExactly("singleName");
			}

			@Test
			void shouldReturnList() {
				var processorNames = vorgangEventListener.getProcessorNames(createCommand(List.of("name", "otherName")));

				assertThat(processorNames).isInstanceOf(Collection.class).containsExactly("name", "otherName");
			}

			private Command createCommand(Object processorNames) {
				var map = new HashMap<String, Object>();
				map.put(ProcessorEventListener.COMMAND_PROCESSOR_NAMES_KEY, processorNames);
				return CommandTestFactory.createBuilder().bodyObject(map).build();
			}
		}

		@DisplayName("publish command processed event")
		@Nested
		class TestPublishCommandProcessedEvent {

			@BeforeEach
			void init() {
				when(processorService.processVorgang(any(), any())).thenReturn(Stream.empty());
			}

			@Test
			void shouldPublishEventBeClassOf() {
				vorgangEventListener.onCommandProcessVorgang(event);

				verify(publisher).publishEvent(commandSuccessEventCaptor.capture());
				assertThat(commandSuccessEventCaptor.getValue().getClass()).isEqualTo(VorgangProcessedEvent.class);
			}

			@Test
			void shouldPublishEventHasCommanIdAsSource() {
				vorgangEventListener.onCommandProcessVorgang(event);

				verify(publisher).publishEvent(commandSuccessEventCaptor.capture());
				assertThat(commandSuccessEventCaptor.getValue().getSource()).isEqualTo(CommandTestFactory.ID);
			}

			@Test
			void shouldPublishEventHasCommand() {
				vorgangEventListener.onCommandProcessVorgang(event);

				verify(publisher).publishEvent(commandSuccessEventCaptor.capture());
				assertThat(commandSuccessEventCaptor.getValue().getCommand()).isEqualTo(command);
			}
		}

		@DisplayName("On ViolationConstraint exception")
		@Nested
		class TestOnViolationConstraintException {

			private final Mono<ProcessorResult> processorResults = Mono.just(ProcessorResultTestFactory.create());
			private final Stream<Mono<ProcessorResult>> monoStream = Stream.of(processorResults);
			private final ConstraintViolationException exception = new ConstraintViolationException("", Collections.emptySet());

			@BeforeEach
			void init() {
				when(vorgangService.getVorgang(any())).thenReturn(vorgang);
				when(processorService.processVorgang(any(), any())).thenReturn(monoStream);
				doThrow(exception).when(vorgangEventListener).processResult(any(), any());
				doNothing().when(vorgangEventListener).publishCommandFailedEvent(any(), any());
			}

			@Test
			void shouldProcessError() {
				vorgangEventListener.onCommandProcessVorgang(event);

				verify(resultService).processError(any(), eq(CommandTestFactory.VORGANG_ID));
			}

			@Test
			void shouldPublishEventOnConstraintException() {
				vorgangEventListener.onCommandProcessVorgang(event);

				verify(vorgangEventListener).publishCommandFailedEvent(any(), any());
			}
		}

		@DisplayName("on process vorgang exception")
		@Nested
		class TestOnProcessVorgangException {

			@BeforeEach
			void init() {
				when(processorService.processVorgang(any(), any())).thenReturn(Stream.of(Mono.error(exception)));
			}

			@Test
			void shouldProcessError() {
				vorgangEventListener.onCommandProcessVorgang(event);

				verify(vorgangEventListener).publishCommandFailedEvent(any(), any());
			}

			@Test
			void shouldPublishCommandFailedEvent() {
				vorgangEventListener.onCommandProcessVorgang(event);

				verify(resultService).processError(any(), eq(CommandTestFactory.VORGANG_ID));
			}
		}

		@DisplayName("publish command failed event")
		@Nested
		class TestPublishCommandFailedEvent {

			private final static String ERROR_MSG = "ErrorMsg";

			@BeforeEach
			void init() {
				when(exception.getMessage()).thenReturn(ERROR_MSG);
			}

			@Test
			void shouldPublishEventBeClassOf() {
				vorgangEventListener.publishCommandFailedEvent(exception, CommandTestFactory.ID);

				verify(publisher).publishEvent(commandFailedEventCaptor.capture());
				assertThat(commandFailedEventCaptor.getValue().getClass()).isEqualTo(CommandFailedEvent.class);
			}

			@Test
			void shouldPublishEventHasCommanIdAsSource() {
				vorgangEventListener.publishCommandFailedEvent(exception, CommandTestFactory.ID);

				verify(publisher).publishEvent(commandFailedEventCaptor.capture());
				assertThat(commandFailedEventCaptor.getValue().getSource()).isEqualTo(CommandTestFactory.ID);
			}

			@Test
			void shouldPublishEventHasErrorMessage() {
				vorgangEventListener.publishCommandFailedEvent(exception, CommandTestFactory.ID);

				verify(publisher).publishEvent(commandFailedEventCaptor.capture());
				assertThat(commandFailedEventCaptor.getValue().getErrorMessage()).isEqualTo(ERROR_MSG);
			}
		}
	}
}